// ***********************************************************
// Use this code at your own risk.
//
// Direct questions and/or fixes to:
//   Illah Nourbakhsh     illah@cs.cmu.edu
//   Kevin Wilson         kwilson@cs.cmu.edu
//   Carnegie Mellon University Robotics Institute
//
// Last updated:  June 4, 1998 (SCOUT compatible)
//
// ***********************************************************

#include "include\interfaceprivate.h"
#include <math.h>

extern SensorType Sensors;
extern CRITICAL_SECTION SerialLock, // serialize serial port accesses
	SonarLock,    // serialize sonar state changes
	CommandLock;  // serialize commands sent to the robot
extern HANDLE hCom1;
extern int SonarsActive, SonarsHidden, SonarActive[], KeybdVisible;
extern HPEN hDrawPen, hErasePen, hKeyDrawPen, hKeyErasePen;
extern HWND hKeybdWnd;

int SonarScaleFactor = 1;
HDC hSonarDC;
HDC hKeybdDC;

// combine to bytes into an integer
int TWOBYTESTOINT(unsigned char low, unsigned char high)
{
	int res;

	res = 256*high + low;
	if (res >= 32768)
		res = 32768 - res;

	return res;	
}

// find occurences of \n in in and convert to \r\n in out.
int n2rn(char *out, char *in)
{
	int i = 0, j = 0, retval = 0;

	while(in[i]) {
		if (in[i] == '\n') {
			out[j] = '\r';
			out[j+1] = '\n';
			j += 2;
			retval += 1;
		} else {
			out[j] = in[i];
			j += 1;
		}
		i += 1;
	}
	out[j] = 0;

	return retval;
}

// find occurences of \r\n in in and convert to \n in out.
int rn2n(char *out, char *in)
{
	int i = 0, j = 0, retval = 0;

	while(in[i]) {
		if ((in[i] == '\r') && (in[i+1] == '\n')) {
			out[j] = '\n';
			i += 2;
			retval += 1;
		} else {
			out[j] = in[i];
			i += 1;
		}
		j += 1;
	}
	out[j] = 0;

	return retval;
}

// append c to packet buf, increment *ptr
void Append2Packet(char *buf, int *ptr, char c)
{
	buf[*ptr] = c;
	*ptr = *ptr + 1;
}

// send a packet buf to the robot and use response to update Sensors.
// timeout is the initial waiting period timeout that is passed to 
// ReadPackage().
// returns:
// 0 on success
// -1 on time out
// -2 on invalid packet start byte
// -3 on checksum error
// -4 on invalid packet end byte
// -5 on invalid packet length
// -6 on error setting serial port parameters
// -20 on send error
//
// packet format is:
// byte 0: 1
// byte 1: low byte of packet length
// byte 2: high byte of packet length
// byte 3: command code
// byte 4 through (buflen-3): command parameters
// byte (buflen-2): checksum
// byte (buflen-1): 92
int RobotCommand(char *buf, int buflen, int timeout)
{
    char pktbuf[50];
	int result;
	
	EnterCriticalSection(&CommandLock);
	PurgeComm(hCom1, PURGE_RXCLEAR);
		
	if (SendPackage(buf, buflen)) {
		LeaveCriticalSection(&CommandLock);
		return -20;
	}

	result = ReadPackage(pktbuf, 100, 50);
	if (result != 0) {
		LeaveCriticalSection(&CommandLock);
		return result;
	}
	
	ProcessResponse(pktbuf, &Sensors);

	LeaveCriticalSection(&CommandLock);
	return 0;
	
}

// send buf to serial port
// returns 0 on success and -1 on failure.  i'd be really surprised if this
// ever failed after successfully opening the serial port.
int SendPackage(char *buf, int buflen)
{
	int i, err;
	
	for(i=0;i<buflen;i++) {
		if (err = SendByte(buf[i])) {
			return err; 
		}
	}

	FlushFileBuffers(hCom1);

	return 0;
}

// sends a byte to the serial port.  returns 0 on success and -1 on failure.
int SendByte(char c)
{
	int numBytes = 2;

	EnterCriticalSection(&SerialLock);	
	if (!WriteFile(hCom1, (void *) &c, 
				   1, &numBytes, NULL) || 
		numBytes != 1) {
		LeaveCriticalSection(&SerialLock);
		return -1;
	}
	
	LeaveCriticalSection(&SerialLock);
	return 0;
}



/*	Note to Illah:
	I changed the semantics of the two timeouts to
	make them more straightforward to implement with
	Win32 reads.  initialTimeout is now how long 
	ReadPackage waits for the first character, and
	auxTimeout is now how long it waits subsequent
	characters.
*/
// reads a reply packet from the robot
// returns:
// 0 on success
// -1 on time out
// -2 on invalid packet start character
// -3 on checksum error
// -4 on invalid packet end character
// -5 on invalid packet length
// -6 on error setting serial port paramters
int ReadPackage(char pktbuf[], int initialTimeout, int auxTimeout)
{
	int sum = 0, index = 0, timeout = 0,
		error = 0, pktlen = 0, i;
	char ch;
	COMMTIMEOUTS timeouts;

	timeouts.ReadIntervalTimeout = 0;
	timeouts.ReadTotalTimeoutMultiplier = initialTimeout;
	timeouts.ReadTotalTimeoutConstant = 0;

	if (!SetCommTimeouts(hCom1, &timeouts))
		return (-6);

	timeout = (ReadByte(&ch) != 0);
	sum += ch;
	if (timeout) 
		return -1;

	if (ch != 1)
		return -2;

	timeouts.ReadIntervalTimeout = 0;
	timeouts.ReadTotalTimeoutMultiplier = auxTimeout;
	timeouts.ReadTotalTimeoutConstant = 0;

	if (!SetCommTimeouts(hCom1, &timeouts))
		return (-6);

	if (ReadByte(&ch) != 0)
		return -1;
	sum += ch;
	pktlen = ch;

	if (ReadByte(&ch) != 0)
		return -1;
	sum += ch;
	pktlen += ch * 256;

	if ((pktlen < 1) || (pktlen > 40))
		return -5;


	for (i=0; i<=pktlen; i++) {
        if (ReadByte(&ch) != 0)
            return -1;
        sum += ch;
        pktbuf[i] = ch;
    }

    if (((sum - 92 - pktbuf[pktlen-1]) & 0xFF) != 
		(unsigned char) pktbuf[pktlen-1])
        return -3;

    if (92 != pktbuf[pktlen])
        return -4;

	return (0);
}

// takes raw packet and fills in a SensorType struct.
void ProcessResponse(char pktbuf[], SensorType *sensorbuf)
{
    int i;

	sensorbuf->lastCommand = pktbuf[0];

    for (i=1; i<=16; i++) {
        sensorbuf->sonar[i-1] = pktbuf[i] & 0xFF;
    }

	sensorbuf->bumper = pktbuf[17] + 
						pktbuf[18] * 256 +
						pktbuf[19] * 65536;

    sensorbuf->xPos = TWOBYTESTOINT(pktbuf[20], pktbuf[21]);
    sensorbuf->yPos = TWOBYTESTOINT(pktbuf[22], pktbuf[23]);
    sensorbuf->angle = TWOBYTESTOINT(pktbuf[24], pktbuf[25]);
	//sensorbuf->turretAngle = 0; // legacy placeholder, 26 & 27 //
    sensorbuf->rVel = TWOBYTESTOINT(pktbuf[28], pktbuf[29]);
    sensorbuf->lVel = TWOBYTESTOINT(pktbuf[30], pktbuf[31]);
	//sensorbuf->turretVel = 0;
    //sensorbuf->turretVel = TWOBYTESTOINT(pktbuf[29], pktbuf[30]);
	sensorbuf->motorStatus = (int)pktbuf[32];
    //sensorbuf->motorStatus = TWOBYTESTOINT(pktbuf[31], pktbuf[32]);

	if (SonarsActive)
		UpdateSonarWindow(sensorbuf);
}

// read byte from serial port.
// returns 0 on success, -1 on failure.
int ReadByte(char *ptr)
{
	int numread;

	EnterCriticalSection(&SerialLock);	
	ReadFile(hCom1, ptr, 1, &numread, NULL);
	LeaveCriticalSection(&SerialLock);

	if (numread != 1) 
		return -1;
	
	return 0;
}

// draws radial lines of appropriate lenght in the sonar display window.
// returns -1 if sonars are off or hidden.  returns 0 otherwise.
int UpdateSonarWindow(SensorType *sensorbuf)
{
	POINT ends[2];
	int i, sonarval, localscale;
	double pi = 3.14159265359L, xmul, ymul, arg, scale;
	static int dirvectors[NUMSONARS][2];
	static int oldsonar[NUMSONARS] = {0};
	HPEN hOldPen;

	localscale = SonarScaleFactor;
	
	if (!SonarsActive || SonarsHidden)
		return -1;
	
	ends[0].x = SONARWINSIZE / 2; 
	ends[0].y = SONARWINSIZE / 2; // center of window

	if (!dirvectors[0][0] && !dirvectors[0][1]) {
		// if this is the first call to UpdateSonarWindow,
		// initialize a table of direction vectors.
		scale = 100.0 * SONARWINSIZE * 0.85 / 2.0 / 255;
		
		for (i=0; i<=NUMSONARS-1; i++) {
			arg = (i * pi * 2.0 / NUMSONARS) + (pi / 2.0);
			xmul = cos(arg) * scale; 
			ymul = -1.0 * sin(arg) * scale;

			dirvectors[i][0] = (int) xmul;		
			dirvectors[i][1] = (int) ymul;
		}
	}

	// erase old lines
	hOldPen = SelectObject(hSonarDC, hErasePen);
	for (i = 0; i <= NUMSONARS - 1; i++) {
		ends[1].x = ends[0].x +
			oldsonar[i] * dirvectors[i][0] / 100;
		ends[1].y = ends[0].y +
			oldsonar[i] * dirvectors[i][1] / 100;
		Polyline(hSonarDC, ends, 2);
	}		

	SelectObject(hSonarDC, hDrawPen);
	for (i = 0; i <= NUMSONARS - 1; i++) {
		if (SonarActive[i]) {
			// draw new line if the sonar is firing
			oldsonar[i] = sonarval =
				MIN(sensorbuf->sonar[i] * localscale, 255);
			ends[1].x = ends[0].x + sonarval * dirvectors[i][0] / 100;
			ends[1].y = ends[0].y + sonarval * dirvectors[i][1] / 100;
			Polyline(hSonarDC, ends, 2);
		} else {
			oldsonar[i] = 0;
		}
	}

	SelectObject(hSonarDC, hOldPen);
	return 0;
}

// update position of crosshairs in keyboard control.
// returns -1 if the keyboard control window isn't visible, 0 otherwise.
int UpdateKeybdWindow(int tv, int rv)
{
	POINT ends[2];
	int xmul = 100 * SONARWINSIZE / 990,
		ymul = 100 * SONARWINSIZE / 550, midx = SONARWINSIZE / 2,
		midy = SONARWINSIZE / 2 + 24 + 2;
	static int oldtv = 0, oldrv = 0;
	HPEN hOldPen;

	if (!KeybdVisible)
		return -1;
	
	hOldPen = SelectObject(hKeybdDC, hKeyErasePen);
	// erase old crosshairs
	ends[0].x = midx - xmul * oldrv / 100 - 4;
	ends[0].y = midy - ymul * oldtv / 100 - 4;
	ends[1].x = midx - xmul * oldrv / 100 + 4;
	ends[1].y = midy - ymul * oldtv / 100 + 4;	
	Polyline(hKeybdDC, ends, 2);

	ends[0].x = midx - xmul * oldrv / 100 - 4;
	ends[0].y = midy - ymul * oldtv / 100 + 4;
	ends[1].x = midx - xmul * oldrv / 100 + 4;
	ends[1].y = midy - ymul * oldtv / 100 - 4;	
	Polyline(hKeybdDC, ends, 2);

	SelectObject(hKeybdDC, hKeyDrawPen);
	// draw new crosshairs
	ends[0].x = midx - xmul * rv / 100 - 3;
	ends[0].y = midy - ymul * tv / 100 - 3;
	ends[1].x = midx - xmul * rv / 100 + 4;
	ends[1].y = midy - ymul * tv / 100 + 4;	
	Polyline(hKeybdDC, ends, 2);

	ends[0].x = midx - xmul * rv / 100 - 3;
	ends[0].y = midy - ymul * tv / 100 + 3;
	ends[1].x = midx - xmul * rv / 100 + 4;
	ends[1].y = midy - ymul * tv / 100 - 4;	
	Polyline(hKeybdDC, ends, 2);

	oldtv = tv;
	oldrv = rv;
	
	SelectObject(hKeybdDC, hOldPen);

	hOldPen = SelectObject(hKeybdDC, hKeyDrawPen);
	// draw axes
	ends[1].x = ends[0].x = SONARWINSIZE / 2;
	ends[0].y = 2 + 24 + 2;
	ends[1].y = SONARWINSIZE + 24 + 2;	
	Polyline(hKeybdDC, ends, 2);

	ends[1].y = ends[0].y = SONARWINSIZE / 2 + 24 + 2;
	ends[0].x = 2;
	ends[1].x = SONARWINSIZE;	
	Polyline(hKeybdDC, ends, 2);

	SelectObject(hKeybdDC, hOldPen);
	
	return 0;
}

// return pointer to the character immediately after the first occurence
// of \n in str.
char *SkipLine(char *str) 
{

	while(1) {
		if (*str == '\n')
			return (str+1);
		else if (*str == '\0')
			return str;
		str++;
	}
}

// hide keyboard window
void HideKeybd()
{
	ShowWindow(hKeybdWnd, SW_HIDE);
	KeybdVisible = 0;
}

// unhide keyboard window
void UnhideKeybd()
{
	ShowWindow(hKeybdWnd, SW_SHOW);
	KeybdVisible = 1;
}

